import type { SpecialCardType } from "./cards"
import type { GameAction } from "./gameActions"
import { produce } from "immer"
import { specialCardTypes } from "./cards"

type SpecialCardCountByType = Record<SpecialCardType, number>

export interface GameStatePlayer {
	id: string
	numberCards: number[]
	specialCardCountByType: SpecialCardCountByType
	stayed: boolean
	nextRoundCovert: boolean
}

export interface GameState {
	players: GameStatePlayer[]
	activeSpecialCardCountByType: SpecialCardCountByType
	activePlayerId: string
	targetSum: number
}

const ALL_ZERO_SPECIAL_CARD_COUNTS: SpecialCardCountByType = Object.fromEntries(specialCardTypes.map(t => [t, 0])) as SpecialCardCountByType

const UNINITIALIZED_GAME_STATE: GameState = {
	players: [],
	activePlayerId: "",
	targetSum: 0,
	activeSpecialCardCountByType: ALL_ZERO_SPECIAL_CARD_COUNTS
}

export const produceNewState = (oldState: GameState, action: GameAction) => produce(oldState, state => {
	switch (action.type) {
		case "start":
			state.players = Object.entries(action.numberCardsByPlayerId).map(([playerId, numberCards]): GameStatePlayer => ({
				id: playerId,
				numberCards,
				specialCardCountByType: ALL_ZERO_SPECIAL_CARD_COUNTS,
				stayed: false,
				nextRoundCovert: false
			}))

			state.activePlayerId = state.players[0].id
			state.targetSum = action.targetSum
			state.activeSpecialCardCountByType = ALL_ZERO_SPECIAL_CARD_COUNTS

			break

		case "deal-number":
			const p1 = state.players.find(p => p.id === action.toPlayerId)!
			p1.numberCards.push(action.number)
			break

		case "deal-special":
			const p2 = state.players.find(p => p.id === action.toPlayerId)!
			p2.specialCardCountByType[action.cardType]++
			break

		case "draw":
		case "stay":
			const activePlayerIndex = state.players.findIndex(p => p.id === state.activePlayerId)

			// activate the next player
			state.activePlayerId = state.players[(activePlayerIndex + 1) % state.players.length].id
			break

		case "use-special":
			const p3 = state.players.find(p => p.id === action.initiatingPlayerId)!
			applySpecialCardUsage(state, action.cardType, p3)
			break
	}
})

function applySpecialCardUsage(state: GameState, type: SpecialCardType, player: GameStatePlayer) {
	player.specialCardCountByType[type]--

	switch (type) {
		case "add-2-opponent":
			// nothing
			break

		case "destroy-last-opponent-special":
			// TODO
			break

		case "double-draw":
			// nothing
			break

		case "increase-target-by-2":
			state.targetSum += 2
			break

		case "next-round-covert":
			player.nextRoundCovert = true
			break

		case "return-last-opponent":
			// TODO
			break

		case "return-last-own":
			// TODO
			break
	}
}